@@ -1,5 +1,38 @@
+/*	$NetBSD: compat.c,v 1.58 2005/05/08 04:19:12 christos Exp $	*/
+
 /*
  * Copyright (c) 1988, 1989, 1990 The Regents of the University of California.
+ * All rights reserved.
+ *
+ * This code is derived from software contributed to Berkeley by
+ * Adam de Boor.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. Neither the name of the University nor the names of its contributors
+ *    may be used to endorse or promote products derived from this software
+ *    without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
+
+/*
  * Copyright (c) 1988, 1989 by Adam de Boor
  * Copyright (c) 1989 by Berkeley Softworks
  * All rights reserved.
@@ -36,9 +69,18 @@
  * SUCH DAMAGE.
  */
 
+#ifndef MAKE_NATIVE
+static char rcsid[] = "$NetBSD: compat.c,v 1.58 2005/05/08 04:19:12 christos Exp $";
+#else
+#include <sys/cdefs.h>
 #ifndef lint
-static char sccsid[] = "@(#)compat.c	5.7 (Berkeley) 3/1/91";
+#if 0
+static char sccsid[] = "@(#)compat.c	8.2 (Berkeley) 3/19/94";
+#else
+__RCSID("$NetBSD: compat.c,v 1.58 2005/05/08 04:19:12 christos Exp $");
+#endif
 #endif /* not lint */
+#endif
 
 /*-
  * compat.c --
@@ -53,14 +95,20 @@ static char sccsid[] = "@(#)compat.c	5.7 (Berkeley) 3/1/91";
  *	    	  	    thems as need creatin'
  */
 
-#include    <stdio.h>
 #include    <sys/types.h>
-#include    <sys/signal.h>
+#include    <sys/stat.h>
 #include    <sys/wait.h>
-#include    <sys/errno.h>
+
 #include    <ctype.h>
+#include    <errno.h>
+#include    <signal.h>
+#include    <stdio.h>
+
 #include    "make.h"
-extern int errno;
+#include    "hash.h"
+#include    "dir.h"
+#include    "job.h"
+#include    "pathnames.h"
 
 /*
  * The following array is used to make a fast determination of which
@@ -73,7 +121,23 @@ static char 	    meta[256];
 
 static GNode	    *curTarg = NILGNODE;
 static GNode	    *ENDNode;
-static int  	    CompatRunCommand();
+static void CompatInterrupt(int);
+
+static void
+Compat_Init(void)
+{
+    const char *cp;
+
+    Shell_Init();		/* setup default shell */
+    
+    for (cp = "#=|^(){};&<>*?[]:$`\\\n"; *cp != '\0'; cp++) {
+	meta[(unsigned char) *cp] = 1;
+    }
+    /*
+     * The null character serves as a sentinel in the string.
+     */
+    meta[0] = 1;
+}
 
 /*-
  *-----------------------------------------------------------------------
@@ -91,17 +155,19 @@ static int  	    CompatRunCommand();
  *-----------------------------------------------------------------------
  */
 static void
-CompatInterrupt (signo)
-    int	    signo;
+CompatInterrupt(int signo)
 {
     GNode   *gn;
-    
-    if ((curTarg != NILGNODE) && !Targ_Precious (curTarg)) {
-	char 	  *file = Var_Value (TARGET, curTarg);
 
-	if (unlink (file) == SUCCESS) {
-	    printf ("*** %s removed\n", file);
+    if ((curTarg != NILGNODE) && !Targ_Precious (curTarg)) {
+	char	  *p1;
+	char 	  *file = Var_Value(TARGET, curTarg, &p1);
+
+	if (!noExecute && eunlink(file) != -1) {
+	    Error("*** %s removed", file);
 	}
+	if (p1)
+	    free(p1);
 
 	/*
 	 * Run .INTERRUPT only if hit with interrupt signal
@@ -109,11 +175,12 @@ CompatInterrupt (signo)
 	if (signo == SIGINT) {
 	    gn = Targ_FindNode(".INTERRUPT", TARG_NOCREATE);
 	    if (gn != NILGNODE) {
-		Lst_ForEach(gn->commands, CompatRunCommand, (ClientData)gn);
+		Compat_Make(gn, gn);
 	    }
 	}
+
     }
-    exit (0);
+    exit (signo);
 }
 
 /*-
@@ -122,6 +189,10 @@ CompatInterrupt (signo)
  *	Execute the next command for a target. If the command returns an
  *	error, the node's made field is set to ERROR and creation stops.
  *
+ * Input:
+ *	cmdp		Command to execute
+ *	gnp		Node from which the command came
+ *
  * Results:
  *	0 if the command succeeded, 1 if an error occurred.
  *
@@ -130,47 +201,58 @@ CompatInterrupt (signo)
  *
  *-----------------------------------------------------------------------
  */
-static int
-CompatRunCommand (cmd, gn)
-    char    	  *cmd;	    	/* Command to execute */
-    GNode   	  *gn;    	/* Node from which the command came */
+int
+CompatRunCommand(ClientData cmdp, ClientData gnp)
 {
     char    	  *cmdStart;	/* Start of expanded command */
-    register char *cp;
+    char 	  *cp, *bp;
     Boolean 	  silent,   	/* Don't print command */
+	    	  doIt,		/* Execute even if -n */
 		  errCheck; 	/* Check errors */
-    union wait 	  reason;   	/* Reason for child's death */
+    int 	  reason;   	/* Reason for child's death */
     int	    	  status;   	/* Description of child's death */
     int	    	  cpid;	    	/* Child actually found */
-    int	    	  numWritten;	/* Number of bytes written for error message */
-    ReturnStatus  stat;	    	/* Status of fork */
+    ReturnStatus  retstat;    	/* Status of fork */
     LstNode 	  cmdNode;  	/* Node where current command is located */
-    char    	  **av;	    	/* Argument vector for thing to exec */
+    const char  **av;	    	/* Argument vector for thing to exec */
     int	    	  argc;	    	/* Number of arguments in av or 0 if not
 				 * dynamically allocated */
     Boolean 	  local;    	/* TRUE if command should be executed
 				 * locally */
-
-    silent = gn->type & OP_SILENT;
-    errCheck = !(gn->type & OP_IGNORE);
-
-    cmdNode = Lst_Member (gn->commands, (ClientData)cmd);
-    cmdStart = Var_Subst (cmd, gn, FALSE);
+    char	  *cmd = (char *) cmdp;
+    GNode	  *gn = (GNode *) gnp;
 
     /*
-     * brk_string will return an argv with a NULL in av[1], thus causing
+     * Avoid clobbered variable warnings by forcing the compiler
+     * to ``unregister'' variables
+     */
+#if __GNUC__
+    (void) &av;
+    (void) &errCheck;
+    (void) &cmd;
+#endif
+    silent = gn->type & OP_SILENT;
+    errCheck = !(gn->type & OP_IGNORE);
+    doIt = FALSE;
+    
+    cmdNode = Lst_Member(gn->commands, (ClientData)cmd);
+    cmdStart = Var_Subst(NULL, cmd, gn, FALSE);
+
+    /*
+     * brk_string will return an argv with a NULL in av[0], thus causing
      * execvp to choke and die horribly. Besides, how can we execute a null
      * command? In any case, we warn the user that the command expanded to
      * nothing (is this the right thing to do?).
      */
-     
+
     if (*cmdStart == '\0') {
+	free(cmdStart);
 	Error("%s expands to empty string", cmd);
 	return(0);
     } else {
 	cmd = cmdStart;
     }
-    Lst_Replace (cmdNode, (ClientData)cmdStart);
+    Lst_Replace(cmdNode, (ClientData)cmdStart);
 
     if ((gn->type & OP_SAVE_CMDS) && (gn != ENDNode)) {
 	(void)Lst_AtEnd(ENDNode->commands, (ClientData)cmdStart);
@@ -180,21 +262,32 @@ CompatRunCommand (cmd, gn)
 	return(0);
     }
 
-    while ((*cmd == '@') || (*cmd == '-')) {
-	if (*cmd == '@') {
+    while ((*cmd == '@') || (*cmd == '-') || (*cmd == '+')) {
+	switch (*cmd) {
+	case '@':
 	    silent = TRUE;
-	} else {
+	    break;
+	case '-':
 	    errCheck = FALSE;
+	    break;
+	case '+':
+	    doIt = TRUE;
+	    if (!meta[0])		/* we came here from jobs */
+		Compat_Init();
+	    break;
 	}
 	cmd++;
     }
-    
+
+    while (isspace((unsigned char)*cmd))
+	cmd++;
+
     /*
      * Search for meta characters in the command. If there are no meta
      * characters, there's no need to execute a shell to execute the
      * command.
      */
-    for (cp = cmd; !meta[*cp]; cp++) {
+    for (cp = cmd; !meta[(unsigned char)*cp]; cp++) {
 	continue;
     }
 
@@ -202,7 +295,7 @@ CompatRunCommand (cmd, gn)
      * Print the command before echoing if we're not supposed to be quiet for
      * this one. We also print the command if -n given.
      */
-    if (!silent || noExecute) {
+    if (!silent || NoExecute(gn)) {
 	printf ("%s\n", cmd);
 	fflush(stdout);
     }
@@ -211,35 +304,38 @@ CompatRunCommand (cmd, gn)
      * If we're not supposed to execute any commands, this is as far as
      * we go...
      */
-    if (noExecute) {
+    if (!doIt && NoExecute(gn)) {
 	return (0);
     }
-    
+
     if (*cp != '\0') {
 	/*
 	 * If *cp isn't the null character, we hit a "meta" character and
-	 * need to pass the command off to the shell. We give the shell the
-	 * -e flag as well as -c if it's supposed to exit when it hits an
-	 * error.
+	 * need to pass the command off to the shell. 
 	 */
-	static char	*shargv[4] = { "/bin/sh" };
+	static const char *shargv[4];
 
-	shargv[1] = (errCheck ? "-ec" : "-c");
+	shargv[0] = shellPath;
+	/*
+	 * The following work for any of the builtin shell specs.
+	 */
+	if (DEBUG(SHELL))
+		shargv[1] = "-xc";
+	else
+		shargv[1] = "-c";
 	shargv[2] = cmd;
 	shargv[3] = (char *)NULL;
 	av = shargv;
 	argc = 0;
+	bp = NULL;
     } else {
 	/*
 	 * No meta-characters, so no need to exec a shell. Break the command
 	 * into words to form an argument vector we can execute.
-	 * brk_string sticks our name in av[0], so we have to
-	 * skip over it...
 	 */
-	av = brk_string(cmd, &argc);
-	av += 1;
+	av = (const char **)brk_string(cmd, &argc, TRUE, &bp);
     }
-    
+
     local = TRUE;
 
     /*
@@ -250,46 +346,60 @@ CompatRunCommand (cmd, gn)
 	Fatal("Could not fork");
     }
     if (cpid == 0) {
-	if (local) {
-	    execvp(av[0], av);
-	    numWritten = write (2, av[0], strlen (av[0]));
-	    numWritten = write (2, ": not found\n", sizeof(": not found"));
-	} else {
-	    (void)execv(av[0], av);
-	}
-	exit(1);
+	Check_Cwd(av);
+	if (local)
+	    (void)execvp(av[0], (char *const *)UNCONST(av));
+	else
+	    (void)execv(av[0], (char *const *)UNCONST(av));
+	execError("exec", av[0]);
+	_exit(1);
     }
-    
+    if (bp) {
+	free(av);
+	free(bp);
+    }
+    Lst_Replace(cmdNode, (ClientData) NULL);
+
     /*
      * The child is off and running. Now all we can do is wait...
      */
     while (1) {
-	int 	  id;
 
-	if (!local) {
-	    id = 0;
-	}
-
-	while ((stat = wait((int *)&reason)) != cpid) {
-	    if (stat == -1 && errno != EINTR) {
+	while ((retstat = wait(&reason)) != cpid) {
+	    if (retstat == -1 && errno != EINTR) {
 		break;
 	    }
 	}
-	
-	if (stat > -1) {
+
+	if (retstat > -1) {
 	    if (WIFSTOPPED(reason)) {
-		status = reason.w_stopval;		/* stopped */
+		status = WSTOPSIG(reason);		/* stopped */
 	    } else if (WIFEXITED(reason)) {
-		status = reason.w_retcode;		/* exited */
+		status = WEXITSTATUS(reason);		/* exited */
 		if (status != 0) {
+		    if (DEBUG(ERROR)) {
+		        printf("\n*** Failed target:  %s\n*** Failed command: ",
+			    gn->name);
+		        for (cp = cmd; *cp; ) {
+    			    if (isspace((unsigned char)*cp)) {
+			        putchar(' ');
+			        while (isspace((unsigned char)*cp))
+				    cp++;
+			    } else {
+			        putchar(*cp);
+			        cp++;
+			    }
+		        }
+			printf ("\n");
+		    }
 		    printf ("*** Error code %d", status);
 		}
 	    } else {
-		status = reason.w_termsig;		/* signaled */
+		status = WTERMSIG(reason);		/* signaled */
 		printf ("*** Signal %d", status);
-	    } 
+	    }
+
 
-	    
 	    if (!WIFEXITED(reason) || (status != 0)) {
 		if (errCheck) {
 		    gn->made = ERROR;
@@ -311,19 +421,24 @@ CompatRunCommand (cmd, gn)
 	    }
 	    break;
 	} else {
-	    Fatal ("error in wait: %d", stat);
+	    Fatal("error in wait: %d: %s", retstat, strerror(errno));
 	    /*NOTREACHED*/
 	}
     }
+    free(cmdStart);
 
     return (status);
 }
 
 /*-
  *-----------------------------------------------------------------------
- * CompatMake --
+ * Compat_Make --
  *	Make a target.
  *
+ * Input:
+ *	gnp		The node to make
+ *	pgnp		Parent to abort if necessary
+ *
  * Results:
  *	0
  *
@@ -332,14 +447,15 @@ CompatRunCommand (cmd, gn)
  *
  *-----------------------------------------------------------------------
  */
-static int
-CompatMake (gn, pgn)
-    GNode   	  *gn;	    /* The node to make */
-    GNode   	  *pgn;	    /* Parent to abort if necessary */
+int
+Compat_Make(ClientData gnp, ClientData pgnp)
 {
-    if (gn->type & OP_USE) {
-	Make_HandleUse(gn, pgn);
-    } else if (gn->made == UNMADE) {
+    GNode *gn = (GNode *) gnp;
+    GNode *pgn = (GNode *) pgnp;
+
+    if (!meta[0])		/* we came here from jobs */
+	Compat_Init();
+    if (gn->made == UNMADE && (gn == pgn || (pgn->type & OP_MADE) == 0)) {
 	/*
 	 * First mark ourselves to be made, then apply whatever transformations
 	 * the suffix module thinks are necessary. Once that's done, we can
@@ -348,20 +464,24 @@ CompatMake (gn, pgn)
 	 * This is our signal to not attempt to do anything but abort our
 	 * parent as well.
 	 */
-	gn->make = TRUE;
+	gn->flags |= REMAKE;
 	gn->made = BEINGMADE;
-	Suff_FindDeps (gn);
-	Lst_ForEach (gn->children, CompatMake, (ClientData)gn);
-	if (!gn->make) {
+	if ((gn->type & OP_MADE) == 0)
+	    Suff_FindDeps(gn);
+	Lst_ForEach(gn->children, Compat_Make, (ClientData)gn);
+	if ((gn->flags & REMAKE) == 0) {
 	    gn->made = ABORTED;
-	    pgn->make = FALSE;
-	    return (0);
+	    pgn->flags &= ~REMAKE;
+	    goto cohorts;
 	}
 
-	if (Lst_Member (gn->iParents, pgn) != NILLNODE) {
-	    Var_Set (IMPSRC, Var_Value(TARGET, gn), pgn);
+	if (Lst_Member(gn->iParents, pgn) != NILLNODE) {
+	    char *p1;
+	    Var_Set(IMPSRC, Var_Value(TARGET, gn, &p1), pgn, 0);
+	    if (p1)
+		free(p1);
 	}
-	
+
 	/*
 	 * All the children were made ok. Now cmtime contains the modification
 	 * time of the newest child, we need to find out if we exist and when
@@ -376,7 +496,7 @@ CompatMake (gn, pgn)
 	    if (DEBUG(MAKE)) {
 		printf("up-to-date.\n");
 	    }
-	    return (0);
+	    goto cohorts;
 	} else if (DEBUG(MAKE)) {
 	    printf("out-of-date.\n");
 	}
@@ -386,7 +506,7 @@ CompatMake (gn, pgn)
 	 * to tell him/her "yes".
 	 */
 	if (queryFlag) {
-	    exit (-1);
+	    exit (1);
 	}
 
 	/*
@@ -395,29 +515,29 @@ CompatMake (gn, pgn)
 	 * Make_DoAllVar().
 	 */
 	Make_DoAllVar(gn);
-		    
+
 	/*
 	 * Alter our type to tell if errors should be ignored or things
 	 * should not be printed so CompatRunCommand knows what to do.
 	 */
-	if (Targ_Ignore (gn)) {
+	if (Targ_Ignore(gn)) {
 	    gn->type |= OP_IGNORE;
 	}
-	if (Targ_Silent (gn)) {
+	if (Targ_Silent(gn)) {
 	    gn->type |= OP_SILENT;
 	}
 
-	if (Job_CheckCommands (gn, Fatal)) {
+	if (Job_CheckCommands(gn, Fatal)) {
 	    /*
 	     * Our commands are ok, but we still have to worry about the -t
 	     * flag...
 	     */
-	    if (!touchFlag) {
+	    if (!touchFlag || (gn->type & OP_MAKE)) {
 		curTarg = gn;
-		Lst_ForEach (gn->commands, CompatRunCommand, (ClientData)gn);
+		Lst_ForEach(gn->commands, CompatRunCommand, (ClientData)gn);
 		curTarg = NILGNODE;
 	    } else {
-		Job_Touch (gn, gn->type & OP_SILENT);
+		Job_Touch(gn, gn->type & OP_SILENT);
 	    }
 	} else {
 	    gn->made = ERROR;
@@ -431,70 +551,15 @@ CompatMake (gn, pgn)
 	     * This is to keep its state from affecting that of its parent.
 	     */
 	    gn->made = MADE;
-#ifndef RECHECK
-	    /*
-	     * We can't re-stat the thing, but we can at least take care of
-	     * rules where a target depends on a source that actually creates
-	     * the target, but only if it has changed, e.g.
-	     *
-	     * parse.h : parse.o
-	     *
-	     * parse.o : parse.y
-	     *  	yacc -d parse.y
-	     *  	cc -c y.tab.c
-	     *  	mv y.tab.o parse.o
-	     *  	cmp -s y.tab.h parse.h || mv y.tab.h parse.h
-	     *
-	     * In this case, if the definitions produced by yacc haven't
-	     * changed from before, parse.h won't have been updated and
-	     * gn->mtime will reflect the current modification time for
-	     * parse.h. This is something of a kludge, I admit, but it's a
-	     * useful one..
-	     *
-	     * XXX: People like to use a rule like
-	     *
-	     * FRC:
-	     *
-	     * To force things that depend on FRC to be made, so we have to
-	     * check for gn->children being empty as well...
-	     */
-	    if (!Lst_IsEmpty(gn->commands) || Lst_IsEmpty(gn->children)) {
-		gn->mtime = now;
-	    }
-#else
-	    /*
-	     * This is what Make does and it's actually a good thing, as it
-	     * allows rules like
-	     *
-	     *	cmp -s y.tab.h parse.h || cp y.tab.h parse.h
-	     *
-	     * to function as intended. Unfortunately, thanks to the stateless
-	     * nature of NFS (and the speed of this program), there are times
-	     * when the modification time of a file created on a remote
-	     * machine will not be modified before the stat() implied by
-	     * the Dir_MTime occurs, thus leading us to believe that the file
-	     * is unchanged, wreaking havoc with files that depend on this one.
-	     *
-	     * I have decided it is better to make too much than to make too
-	     * little, so this stuff is commented out unless you're sure it's
-	     * ok.
-	     * -- ardeb 1/12/88
-	     */
-	    if (noExecute || Dir_MTime(gn) == 0) {
-		gn->mtime = now;
-	    }
-	    if (DEBUG(MAKE)) {
-		printf("update time: %s\n", Targ_FmtTime(gn->mtime));
-	    }
-#endif
+	    pgn->flags |= Make_Recheck(gn) == 0 ? FORCE : 0;
 	    if (!(gn->type & OP_EXEC)) {
-		pgn->childMade = TRUE;
+		pgn->flags |= CHILDMADE;
 		Make_TimeStamp(pgn, gn);
 	    }
 	} else if (keepgoing) {
-	    pgn->make = FALSE;
+	    pgn->flags &= ~REMAKE;
 	} else {
-	    printf ("\n\nStop.\n");
+	    PrintOnError("\n\nStop.");
 	    exit (1);
 	}
     } else if (gn->made == ERROR) {
@@ -502,20 +567,23 @@ CompatMake (gn, pgn)
 	 * Already had an error when making this beastie. Tell the parent
 	 * to abort.
 	 */
-	pgn->make = FALSE;
+	pgn->flags &= ~REMAKE;
     } else {
-	if (Lst_Member (gn->iParents, pgn) != NILLNODE) {
-	    Var_Set (IMPSRC, Var_Value(TARGET, gn), pgn);
+	if (Lst_Member(gn->iParents, pgn) != NILLNODE) {
+	    char *p1;
+	    Var_Set(IMPSRC, Var_Value(TARGET, gn, &p1), pgn, 0);
+	    if (p1)
+		free(p1);
 	}
 	switch(gn->made) {
 	    case BEINGMADE:
-		Error("Graph cycles through %s\n", gn->name);
+		Error("Graph cycles through %s", gn->name);
 		gn->made = ERROR;
-		pgn->make = FALSE;
+		pgn->flags &= ~REMAKE;
 		break;
 	    case MADE:
 		if ((gn->type & OP_EXEC) == 0) {
-		    pgn->childMade = TRUE;
+		    pgn->flags |= CHILDMADE;
 		    Make_TimeStamp(pgn, gn);
 		}
 		break;
@@ -524,17 +592,24 @@ CompatMake (gn, pgn)
 		    Make_TimeStamp(pgn, gn);
 		}
 		break;
+	    default:
+		break;
 	}
     }
 
+cohorts:
+    Lst_ForEach(gn->cohorts, Compat_Make, pgnp);
     return (0);
 }
-	
+
 /*-
  *-----------------------------------------------------------------------
  * Compat_Run --
  *	Initialize this mode and start making.
  *
+ * Input:
+ *	targs		List of target nodes to re-create
+ *
  * Results:
  *	None.
  *
@@ -544,13 +619,13 @@ CompatMake (gn, pgn)
  *-----------------------------------------------------------------------
  */
 void
-Compat_Run(targs)
-    Lst	    	  targs;    /* List of target nodes to re-create */
+Compat_Run(Lst targs)
 {
-    char    	  *cp;	    /* Pointer to string of shell meta-characters */
-    GNode   	  *gn;	    /* Current root target */
+    GNode   	  *gn = NULL;/* Current root target */
     int	    	  errors;   /* Number of targets not remade due to errors */
 
+    Compat_Init();
+
     if (signal(SIGINT, SIG_IGN) != SIG_IGN) {
 	signal(SIGINT, CompatInterrupt);
     }
@@ -564,15 +639,8 @@ Compat_Run(targs)
 	signal(SIGQUIT, CompatInterrupt);
     }
 
-    for (cp = "#=|^(){};&<>*?[]:$`\\\n"; *cp != '\0'; cp++) {
-	meta[*cp] = 1;
-    }
-    /*
-     * The null character serves as a sentinel in the string.
-     */
-    meta[0] = 1;
-
     ENDNode = Targ_FindNode(".END", TARG_CREATE);
+    ENDNode->type = OP_SPECIAL;
     /*
      * If the user has defined a .BEGIN target, execute the commands attached
      * to it.
@@ -580,13 +648,23 @@ Compat_Run(targs)
     if (!queryFlag) {
 	gn = Targ_FindNode(".BEGIN", TARG_NOCREATE);
 	if (gn != NILGNODE) {
-	    Lst_ForEach(gn->commands, CompatRunCommand, (ClientData)gn);
+	    Compat_Make(gn, gn);
+            if (gn->made == ERROR) {
+                PrintOnError("\n\nStop.");
+                exit(1);
+            }
 	}
     }
 
     /*
-     * For each entry in the list of targets to create, call CompatMake on
-     * it to create the thing. CompatMake will leave the 'made' field of gn
+     * Expand .USE nodes right now, because they can modify the structure
+     * of the tree.
+     */
+    Lst_Destroy(Make_ExpandUse(targs), NOFREE);
+
+    /*
+     * For each entry in the list of targets to create, call Compat_Make on
+     * it to create the thing. Compat_Make will leave the 'made' field of gn
      * in one of several states:
      *	    UPTODATE	    gn was already up-to-date
      *	    MADE  	    gn was recreated successfully
@@ -596,8 +674,8 @@ Compat_Run(targs)
      */
     errors = 0;
     while (!Lst_IsEmpty (targs)) {
-	gn = (GNode *) Lst_DeQueue (targs);
-	CompatMake (gn, gn);
+	gn = (GNode *) Lst_DeQueue(targs);
+	Compat_Make(gn, gn);
 
 	if (gn->made == UPTODATE) {
 	    printf ("`%s' is up to date.\n", gn->name);
@@ -611,6 +689,10 @@ Compat_Run(targs)
      * If the user has defined a .END target, run its commands.
      */
     if (errors == 0) {
-	Lst_ForEach(ENDNode->commands, CompatRunCommand, (ClientData)gn);
+	Compat_Make(ENDNode, ENDNode);
+	if (gn->made == ERROR) {
+	    PrintOnError("\n\nStop.");
+	    exit(1);
+	}
     }
 }
