@@ -1,6 +1,38 @@
+/*	$NetBSD: targ.c,v 1.34 2005/02/16 15:11:53 christos Exp $	*/
+
+/*
+ * Copyright (c) 1988, 1989, 1990, 1993
+ *	The Regents of the University of California.  All rights reserved.
+ *
+ * This code is derived from software contributed to Berkeley by
+ * Adam de Boor.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. Neither the name of the University nor the names of its contributors
+ *    may be used to endorse or promote products derived from this software
+ *    without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ */
+
 /*
- * Copyright (c) 1988, 1989, 1990 The Regents of the University of California.
- * Copyright (c) 1988, 1989 by Adam de Boor
  * Copyright (c) 1989 by Berkeley Softworks
  * All rights reserved.
  *
@@ -36,9 +68,18 @@
  * SUCH DAMAGE.
  */
 
+#ifndef MAKE_NATIVE
+static char rcsid[] = "$NetBSD: targ.c,v 1.34 2005/02/16 15:11:53 christos Exp $";
+#else
+#include <sys/cdefs.h>
 #ifndef lint
-static char sccsid[] = "@(#)targ.c	5.9 (Berkeley) 3/1/91";
+#if 0
+static char sccsid[] = "@(#)targ.c	8.2 (Berkeley) 3/19/94";
+#else
+__RCSID("$NetBSD: targ.c,v 1.34 2005/02/16 15:11:53 christos Exp $");
+#endif
 #endif /* not lint */
+#endif
 
 /*-
  * targ.c --
@@ -49,6 +90,10 @@ static char sccsid[] = "@(#)targ.c	5.9 (Berkeley) 3/1/91";
  * Interface:
  *	Targ_Init 	    	Initialization procedure.
  *
+ *	Targ_End 	    	Cleanup the module
+ *
+ *	Targ_List 	    	Return the list of all targets so far.
+ *
  *	Targ_NewGN	    	Create a new GNode for the passed target
  *	    	  	    	(string). The node is *not* placed in the
  *	    	  	    	hash table, though all its fields are
@@ -81,14 +126,28 @@ static char sccsid[] = "@(#)targ.c	5.9 (Berkeley) 3/1/91";
 
 #include	  <stdio.h>
 #include	  <time.h>
+
 #include	  "make.h"
 #include	  "hash.h"
+#include	  "dir.h"
 
 static Lst        allTargets;	/* the list of all targets found so far */
+#ifdef CLEANUP
+static Lst	  allGNs;	/* List of all the GNodes */
+#endif
 static Hash_Table targets;	/* a hash table of same */
 
 #define HTSIZE	191		/* initial size of hash table */
 
+static int TargPrintOnlySrc(ClientData, ClientData);
+static int TargPrintName(ClientData, ClientData);
+static int TargPrintNode(ClientData, ClientData);
+#ifdef CLEANUP
+static void TargFreeGN(ClientData);
+#endif
+static int TargPropagateCohort(ClientData, ClientData);
+static int TargPropagateNode(ClientData, ClientData);
+
 /*-
  *-----------------------------------------------------------------------
  * Targ_Init --
@@ -102,10 +161,51 @@ static Hash_Table targets;	/* a hash table of same */
  *-----------------------------------------------------------------------
  */
 void
-Targ_Init ()
+Targ_Init(void)
 {
-    allTargets = Lst_Init (FALSE);
-    Hash_InitTable (&targets, HTSIZE);
+    allTargets = Lst_Init(FALSE);
+    Hash_InitTable(&targets, HTSIZE);
+}
+
+/*-
+ *-----------------------------------------------------------------------
+ * Targ_End --
+ *	Finalize this module
+ *
+ * Results:
+ *	None
+ *
+ * Side Effects:
+ *	All lists and gnodes are cleared
+ *-----------------------------------------------------------------------
+ */
+void
+Targ_End(void)
+{
+#ifdef CLEANUP
+    Lst_Destroy(allTargets, NOFREE);
+    if (allGNs)
+	Lst_Destroy(allGNs, TargFreeGN);
+    Hash_DeleteTable(&targets);
+#endif
+}
+
+/*-
+ *-----------------------------------------------------------------------
+ * Targ_List --
+ *	Return the list of all targets
+ *
+ * Results:
+ *	The list of all targets.
+ *
+ * Side Effects:
+ *	None
+ *-----------------------------------------------------------------------
+ */
+Lst
+Targ_List(void)
+{
+    return allTargets;
 }
 
 /*-
@@ -113,50 +213,109 @@ Targ_Init ()
  * Targ_NewGN  --
  *	Create and initialize a new graph node
  *
+ * Input:
+ *	name		the name to stick in the new node
+ *
  * Results:
  *	An initialized graph node with the name field filled with a copy
  *	of the passed name
  *
  * Side Effects:
- *	None.
+ *	The gnode is added to the list of all gnodes.
  *-----------------------------------------------------------------------
  */
 GNode *
-Targ_NewGN (name)
-    char           *name;	/* the name to stick in the new node */
+Targ_NewGN(const char *name)
 {
-    register GNode *gn;
+    GNode *gn;
 
-    gn = (GNode *) emalloc (sizeof (GNode));
-    gn->name = strdup (name);
-    gn->path = (char *) 0;
+    gn = emalloc(sizeof(GNode));
+    gn->name = estrdup(name);
+    gn->uname = NULL;
+    gn->path = NULL;
     if (name[0] == '-' && name[1] == 'l') {
 	gn->type = OP_LIB;
     } else {
 	gn->type = 0;
     }
     gn->unmade =    	0;
-    gn->make = 	    	FALSE;
+    gn->unmade_cohorts = 0;
+    gn->centurion =    	NULL;
     gn->made = 	    	UNMADE;
-    gn->childMade = 	FALSE;
+    gn->flags = 	0;
+    gn->order =		0;
     gn->mtime = gn->cmtime = 0;
-    gn->iParents =  	Lst_Init (FALSE);
-    gn->cohorts =   	Lst_Init (FALSE);
-    gn->parents =   	Lst_Init (FALSE);
-    gn->children =  	Lst_Init (FALSE);
+    gn->iParents =  	Lst_Init(FALSE);
+    gn->cohorts =   	Lst_Init(FALSE);
+    gn->parents =   	Lst_Init(FALSE);
+    gn->children =  	Lst_Init(FALSE);
     gn->successors = 	Lst_Init(FALSE);
     gn->preds =     	Lst_Init(FALSE);
-    gn->context =   	Lst_Init (FALSE);
-    gn->commands =  	Lst_Init (FALSE);
+    Hash_InitTable(&gn->context, 0);
+    gn->commands =  	Lst_Init(FALSE);
+    gn->suffix =	NULL;
+    gn->lineno =	0;
+    gn->fname = 	NULL;
+
+#ifdef CLEANUP
+    if (allGNs == NULL)
+	allGNs = Lst_Init(FALSE);
+    Lst_AtEnd(allGNs, (ClientData) gn);
+#endif
 
     return (gn);
 }
 
+#ifdef CLEANUP
+/*-
+ *-----------------------------------------------------------------------
+ * TargFreeGN  --
+ *	Destroy a GNode
+ *
+ * Results:
+ *	None.
+ *
+ * Side Effects:
+ *	None.
+ *-----------------------------------------------------------------------
+ */
+static void
+TargFreeGN(ClientData gnp)
+{
+    GNode *gn = (GNode *) gnp;
+
+
+    free(gn->name);
+    if (gn->uname)
+	free(gn->uname);
+    if (gn->path)
+	free(gn->path);
+    if (gn->fname)
+	free(gn->fname);
+
+    Lst_Destroy(gn->iParents, NOFREE);
+    Lst_Destroy(gn->cohorts, NOFREE);
+    Lst_Destroy(gn->parents, NOFREE);
+    Lst_Destroy(gn->children, NOFREE);
+    Lst_Destroy(gn->successors, NOFREE);
+    Lst_Destroy(gn->preds, NOFREE);
+    Hash_DeleteTable(&gn->context);
+    Lst_Destroy(gn->commands, NOFREE);
+    free((Address)gn);
+}
+#endif
+
+
 /*-
  *-----------------------------------------------------------------------
  * Targ_FindNode  --
  *	Find a node in the list using the given name for matching
  *
+ * Input:
+ *	name		the name to find
+ *	flags		flags governing events when target not
+ *			found
+ *
  * Results:
  *	The node in the list if it was. If it wasn't, return NILGNODE of
  *	flags was TARG_NOCREATE or the newly created and initialized node
@@ -167,10 +326,7 @@ Targ_NewGN (name)
  *-----------------------------------------------------------------------
  */
 GNode *
-Targ_FindNode (name, flags)
-    char           *name;	/* the name to find */
-    int             flags;	/* flags governing events when target not
-				 * found */
+Targ_FindNode(const char *name, int flags)
 {
     GNode         *gn;	      /* node in that element */
     Hash_Entry	  *he;	      /* New or used hash entry for node */
@@ -179,14 +335,15 @@ Targ_FindNode (name, flags)
 
 
     if (flags & TARG_CREATE) {
-	he = Hash_CreateEntry (&targets, name, &isNew);
+	he = Hash_CreateEntry(&targets, name, &isNew);
 	if (isNew) {
-	    gn = Targ_NewGN (name);
+	    gn = Targ_NewGN(name);
 	    Hash_SetValue (he, gn);
-	    (void) Lst_AtEnd (allTargets, (ClientData)gn);
+	    Var_Append(".ALLTARGETS", name, VAR_GLOBAL);
+	    (void) Lst_AtEnd(allTargets, (ClientData)gn);
 	}
     } else {
-	he = Hash_FindEntry (&targets, name);
+	he = Hash_FindEntry(&targets, name);
     }
 
     if (he == (Hash_Entry *) NULL) {
@@ -199,11 +356,15 @@ Targ_FindNode (name, flags)
 /*-
  *-----------------------------------------------------------------------
  * Targ_FindList --
- *	Make a complete list of GNodes from the given list of names 
+ *	Make a complete list of GNodes from the given list of names
+ *
+ * Input:
+ *	name		list of names to find
+ *	flags		flags used if no node is found for a given name
  *
  * Results:
  *	A complete list of graph nodes corresponding to all instances of all
- *	the names in names. 
+ *	the names in names.
  *
  * Side Effects:
  *	If flags is TARG_CREATE, nodes will be created for all names in
@@ -212,39 +373,33 @@ Targ_FindNode (name, flags)
  * -----------------------------------------------------------------------
  */
 Lst
-Targ_FindList (names, flags)
-    Lst        	   names;	/* list of names to find */
-    int            flags;	/* flags used if no node is found for a given
-				 * name */
+Targ_FindList(Lst names, int flags)
 {
     Lst            nodes;	/* result list */
-    register LstNode  ln;		/* name list element */
-    register GNode *gn;		/* node in tLn */
-    char    	  *name;
+    LstNode	   ln;		/* name list element */
+    GNode	   *gn;		/* node in tLn */
+    char    	   *name;
 
-    nodes = Lst_Init (FALSE);
+    nodes = Lst_Init(FALSE);
 
-    if (Lst_Open (names) == FAILURE) {
+    if (Lst_Open(names) == FAILURE) {
 	return (nodes);
     }
-    while ((ln = Lst_Next (names)) != NILLNODE) {
+    while ((ln = Lst_Next(names)) != NILLNODE) {
 	name = (char *)Lst_Datum(ln);
-	gn = Targ_FindNode (name, flags);
+	gn = Targ_FindNode(name, flags);
 	if (gn != NILGNODE) {
 	    /*
 	     * Note: Lst_AtEnd must come before the Lst_Concat so the nodes
 	     * are added to the list in the order in which they were
 	     * encountered in the makefile.
 	     */
-	    (void) Lst_AtEnd (nodes, (ClientData)gn);
-	    if (gn->type & OP_DOUBLEDEP) {
-		(void)Lst_Concat (nodes, gn->cohorts, LST_CONCNEW);
-	    }
+	    (void) Lst_AtEnd(nodes, (ClientData)gn);
 	} else if (flags == TARG_NOCREATE) {
-	    Error ("\"%s\" -- target unknown.", name);
+	    Error("\"%s\" -- target unknown.", name);
 	}
     }
-    Lst_Close (names);
+    Lst_Close(names);
     return (nodes);
 }
 
@@ -253,6 +408,9 @@ Targ_FindList (names, flags)
  * Targ_Ignore  --
  *	Return true if should ignore errors when creating gn
  *
+ * Input:
+ *	gn		node to check for
+ *
  * Results:
  *	TRUE if should ignore errors
  *
@@ -261,8 +419,7 @@ Targ_FindList (names, flags)
  *-----------------------------------------------------------------------
  */
 Boolean
-Targ_Ignore (gn)
-    GNode          *gn;		/* node to check for */
+Targ_Ignore(GNode *gn)
 {
     if (ignoreErrors || gn->type & OP_IGNORE) {
 	return (TRUE);
@@ -276,6 +433,9 @@ Targ_Ignore (gn)
  * Targ_Silent  --
  *	Return true if be silent when creating gn
  *
+ * Input:
+ *	gn		node to check for
+ *
  * Results:
  *	TRUE if should be silent
  *
@@ -284,8 +444,7 @@ Targ_Ignore (gn)
  *-----------------------------------------------------------------------
  */
 Boolean
-Targ_Silent (gn)
-    GNode          *gn;		/* node to check for */
+Targ_Silent(GNode *gn)
 {
     if (beSilent || gn->type & OP_SILENT) {
 	return (TRUE);
@@ -299,6 +458,9 @@ Targ_Silent (gn)
  * Targ_Precious --
  *	See if the given target is precious
  *
+ * Input:
+ *	gn		the node to check
+ *
  * Results:
  *	TRUE if it is precious. FALSE otherwise
  *
@@ -307,8 +469,7 @@ Targ_Silent (gn)
  *-----------------------------------------------------------------------
  */
 Boolean
-Targ_Precious (gn)
-    GNode          *gn;		/* the node to check */
+Targ_Precious(GNode *gn)
 {
     if (allPrecious || (gn->type & (OP_PRECIOUS|OP_DOUBLEDEP))) {
 	return (TRUE);
@@ -320,12 +481,15 @@ Targ_Precious (gn)
 /******************* DEBUG INFO PRINTING ****************/
 
 static GNode	  *mainTarg;	/* the main target, as set by Targ_SetMain */
-/*- 
+/*-
  *-----------------------------------------------------------------------
  * Targ_SetMain --
  *	Set our idea of the main target we'll be creating. Used for
  *	debugging output.
  *
+ * Input:
+ *	gn		The main target we'll create
+ *
  * Results:
  *	None.
  *
@@ -334,17 +498,15 @@ static GNode	  *mainTarg;	/* the main target, as set by Targ_SetMain */
  *-----------------------------------------------------------------------
  */
 void
-Targ_SetMain (gn)
-    GNode   *gn;  	/* The main target we'll create */
+Targ_SetMain(GNode *gn)
 {
     mainTarg = gn;
 }
 
 static int
-TargPrintName (gn, ppath)
-    GNode          *gn;
-    int		    ppath;
+TargPrintName(ClientData gnp, ClientData ppath)
 {
+    GNode *gn = (GNode *) gnp;
     printf ("%s ", gn->name);
 #ifdef notdef
     if (ppath) {
@@ -355,17 +517,16 @@ TargPrintName (gn, ppath)
 	    printf ("(MAIN NAME)  ");
 	}
     }
-#endif notdef
-    return (0);
+#endif /* notdef */
+    return (ppath ? 0 : 0);
 }
 
 
 int
-Targ_PrintCmd (cmd)
-    char           *cmd;
+Targ_PrintCmd(ClientData cmd, ClientData dummy)
 {
-    printf ("\t%s\n", cmd);
-    return (0);
+    printf ("\t%s\n", (char *) cmd);
+    return (dummy ? 0 : 0);
 }
 
 /*-
@@ -383,24 +544,16 @@ Targ_PrintCmd (cmd)
  *-----------------------------------------------------------------------
  */
 char *
-Targ_FmtTime (time)
-    time_t    time;
+Targ_FmtTime(time_t tm)
 {
     struct tm	  	*parts;
-    static char	  	buf[40];
-    static char	  	*months[] = {
-	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
-	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
-    };
+    static char	  	buf[128];
 
-    parts = localtime(&time);
-
-    sprintf (buf, "%d:%02d:%02d %s %d, 19%d",
-	     parts->tm_hour, parts->tm_min, parts->tm_sec,
-	     months[parts->tm_mon], parts->tm_mday, parts->tm_year);
+    parts = localtime(&tm);
+    (void)strftime(buf, sizeof buf, "%k:%M:%S %b %d, %Y", parts);
     return(buf);
 }
-    
+
 /*-
  *-----------------------------------------------------------------------
  * Targ_PrintType --
@@ -414,18 +567,12 @@ Targ_FmtTime (time)
  *-----------------------------------------------------------------------
  */
 void
-Targ_PrintType (type)
-    register int    type;
+Targ_PrintType(int type)
 {
-    register int    tbit;
-    
-#ifdef __STDC__
+    int    tbit;
+
 #define PRINTBIT(attr)	case CONCAT(OP_,attr): printf("." #attr " "); break
 #define PRINTDBIT(attr) case CONCAT(OP_,attr): if (DEBUG(TARG)) printf("." #attr " "); break
-#else
-#define PRINTBIT(attr) 	case CONCAT(OP_,attr): printf(".attr "); break
-#define PRINTDBIT(attr)	case CONCAT(OP_,attr): if (DEBUG(TARG)) printf(".attr "); break
-#endif /* __STDC__ */
 
     type &= ~OP_OPMASK;
 
@@ -448,6 +595,8 @@ Targ_PrintType (type)
 	    /*XXX: MEMBER is defined, so CONCAT(OP_,MEMBER) gives OP_"%" */
 	    case OP_MEMBER: if (DEBUG(TARG)) printf(".MEMBER "); break;
 	    PRINTDBIT(ARCHV);
+	    PRINTDBIT(MADE);
+	    PRINTDBIT(PHONY);
 	}
     }
 }
@@ -459,10 +608,10 @@ Targ_PrintType (type)
  *-----------------------------------------------------------------------
  */
 static int
-TargPrintNode (gn, pass)
-    GNode         *gn;
-    int	    	  pass;
+TargPrintNode(ClientData gnp, ClientData passp)
 {
+    GNode         *gn = (GNode *) gnp;
+    int	    	  pass = *(int *) passp;
     if (!OP_NOP(gn->type)) {
 	printf("#\n");
 	if (gn == mainTarg) {
@@ -474,7 +623,7 @@ TargPrintNode (gn, pass)
 	    } else {
 		printf("# No unmade children\n");
 	    }
-	    if (! (gn->type & (OP_JOIN|OP_USE|OP_EXEC))) {
+	    if (! (gn->type & (OP_JOIN|OP_USE|OP_USEBEFORE|OP_EXEC))) {
 		if (gn->mtime != 0) {
 		    printf("# last modified %s: %s\n",
 			      Targ_FmtTime(gn->mtime),
@@ -494,16 +643,16 @@ TargPrintNode (gn, pass)
 	    }
 	    if (!Lst_IsEmpty (gn->iParents)) {
 		printf("# implicit parents: ");
-		Lst_ForEach (gn->iParents, TargPrintName, (ClientData)0);
-		putc ('\n', stdout);
+		Lst_ForEach(gn->iParents, TargPrintName, (ClientData)0);
+		fputc ('\n', stdout);
 	    }
 	}
 	if (!Lst_IsEmpty (gn->parents)) {
 	    printf("# parents: ");
-	    Lst_ForEach (gn->parents, TargPrintName, (ClientData)0);
-	    putc ('\n', stdout);
+	    Lst_ForEach(gn->parents, TargPrintName, (ClientData)0);
+	    fputc ('\n', stdout);
 	}
-	
+
 	printf("%-16s", gn->name);
 	switch (gn->type & OP_OPMASK) {
 	    case OP_DEPENDS:
@@ -513,13 +662,13 @@ TargPrintNode (gn, pass)
 	    case OP_DOUBLEDEP:
 		printf(":: "); break;
 	}
-	Targ_PrintType (gn->type);
-	Lst_ForEach (gn->children, TargPrintName, (ClientData)0);
-	putc ('\n', stdout);
-	Lst_ForEach (gn->commands, Targ_PrintCmd, (ClientData)0);
+	Targ_PrintType(gn->type);
+	Lst_ForEach(gn->children, TargPrintName, (ClientData)0);
+	fputc ('\n', stdout);
+	Lst_ForEach(gn->commands, Targ_PrintCmd, (ClientData)0);
 	printf("\n\n");
 	if (gn->type & OP_DOUBLEDEP) {
-	    Lst_ForEach (gn->cohorts, TargPrintNode, (ClientData)pass);
+	    Lst_ForEach(gn->cohorts, TargPrintNode, (ClientData)&pass);
 	}
     }
     return (0);
@@ -534,19 +683,18 @@ TargPrintNode (gn, pass)
  *	0.
  *
  * Side Effects:
- *	The name of each file is printed preceeded by #\t
+ *	The name of each file is printed preceded by #\t
  *
  *-----------------------------------------------------------------------
  */
 static int
-TargPrintOnlySrc(gn)
-    GNode   	  *gn;
+TargPrintOnlySrc(ClientData gnp, ClientData dummy)
 {
-    if (OP_NOP(gn->type)) {
-	printf("#\t%s [%s]\n", gn->name,
-		  gn->path ? gn->path : gn->name);
-    }
-    return (0);
+    GNode   	  *gn = (GNode *) gnp;
+    if (OP_NOP(gn->type))
+	printf("#\t%s [%s]\n", gn->name, gn->path ? gn->path : gn->name);
+
+    return (dummy ? 0 : 0);
 }
 
 /*-
@@ -554,6 +702,10 @@ TargPrintOnlySrc(gn)
  * Targ_PrintGraph --
  *	print the entire graph. heh heh
  *
+ * Input:
+ *	pass		Which pass this is. 1 => no processing
+ *			2 => processing done
+ *
  * Results:
  *	none
  *
@@ -561,21 +713,45 @@ TargPrintOnlySrc(gn)
  *	lots o' output
  *-----------------------------------------------------------------------
  */
-Targ_PrintGraph (pass)
-    int	    pass; 	/* Which pass this is. 1 => no processing
-			 * 2 => processing done */
+void
+Targ_PrintGraph(int pass)
 {
     printf("#*** Input graph:\n");
-    Lst_ForEach (allTargets, TargPrintNode, (ClientData)pass);
+    Lst_ForEach(allTargets, TargPrintNode, (ClientData)&pass);
     printf("\n\n");
     printf("#\n#   Files that are only sources:\n");
-    Lst_ForEach (allTargets, TargPrintOnlySrc);
+    Lst_ForEach(allTargets, TargPrintOnlySrc, (ClientData) 0);
     printf("#*** Global Variables:\n");
-    Var_Dump (VAR_GLOBAL);
+    Var_Dump(VAR_GLOBAL);
     printf("#*** Command-line Variables:\n");
-    Var_Dump (VAR_CMD);
+    Var_Dump(VAR_CMD);
     printf("\n");
     Dir_PrintDirectories();
     printf("\n");
     Suff_PrintAll();
 }
+
+static int
+TargPropagateCohort(ClientData cgnp, ClientData pgnp)
+{
+    GNode	  *cgn = (GNode *) cgnp;
+    GNode	  *pgn = (GNode *) pgnp;
+
+    cgn->type |= pgn->type & ~OP_OPMASK;
+    return (0);
+}
+
+static int
+TargPropagateNode(ClientData gnp, ClientData junk __unused)
+{
+    GNode	  *gn = (GNode *) gnp;
+    if (gn->type & OP_DOUBLEDEP)
+	Lst_ForEach(gn->cohorts, TargPropagateCohort, gnp);
+    return (0);
+}
+
+void
+Targ_Propagate(void)
+{
+    Lst_ForEach(allTargets, TargPropagateNode, (ClientData)0);
+}
